import json
import random
import asyncio
import logging
import subprocess


class DNSBrute:
    """
    Helper for DNS brute-forcing.

    Examples:
    >>> domain = "evilcorp.com"
    >>> subdomains = ["www", "mail"]
    >>> results = await self.helpers.dns.brute(self, domain, subdomains)
    """

    _nameservers_url = (
        "https://raw.githubusercontent.com/blacklanternsecurity/public-dns-servers/master/nameservers.txt"
    )

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.log = logging.getLogger("bbot.helper.dns.brute")
        self.dns_config = self.parent_helper.config.get("dns", {})
        self.num_canaries = 100
        self.max_resolvers = self.dns_config.get("brute_threads", 1000)
        self.nameservers_url = self.dns_config.get("brute_nameservers", self._nameservers_url)
        self.devops_mutations = list(self.parent_helper.word_cloud.devops_mutations)
        self.digit_regex = self.parent_helper.re.compile(r"\d+")
        self._resolver_file = None
        self._dnsbrute_lock = asyncio.Lock()

    async def __call__(self, *args, **kwargs):
        return await self.dnsbrute(*args, **kwargs)

    async def dnsbrute(self, module, domain, subdomains, type=None):
        subdomains = list(subdomains)

        if type is None:
            type = "A"
        type = str(type).strip().upper()

        wildcard_rdtypes = await self.parent_helper.dns.is_wildcard_domain(domain, (type, "CNAME"))
        if wildcard_rdtypes:
            self.log.hugewarning(
                f"Aborting massdns on {domain} because it's a wildcard domain ({','.join(wildcard_rdtypes)})"
            )
            return []

        canaries = self.gen_random_subdomains(self.num_canaries)
        canaries_list = list(canaries)
        canaries_pre = canaries_list[: int(self.num_canaries / 2)]
        canaries_post = canaries_list[int(self.num_canaries / 2) :]
        # sandwich subdomains between canaries
        subdomains = canaries_pre + subdomains + canaries_post

        results = []
        canaries_triggered = []
        async for hostname, ip, rdtype in self._massdns(module, domain, subdomains, rdtype=type):
            sub = hostname.split(domain)[0]
            if sub in canaries:
                canaries_triggered.append(sub)
            else:
                results.append(hostname)

        if len(canaries_triggered) > 5:
            self.log.info(
                f"Aborting massdns on {domain} due to false positive: ({len(canaries_triggered):,} canaries triggered - {','.join(canaries_triggered)})"
            )
            return []

        # everything checks out
        return results

    async def _massdns(self, module, domain, subdomains, rdtype):
        """
        {
            "name": "www.blacklanternsecurity.com.",
            "type": "A",
            "class": "IN",
            "status": "NOERROR",
            "data": {
            "answers": [
                {
                "ttl": 3600,
                "type": "CNAME",
                "class": "IN",
                "name": "www.blacklanternsecurity.com.",
                "data": "blacklanternsecurity.github.io."
                },
                {
                "ttl": 3600,
                "type": "A",
                "class": "IN",
                "name": "blacklanternsecurity.github.io.",
                "data": "185.199.108.153"
                }
            ]
            },
            "resolver": "168.215.165.186:53"
        }
        """
        resolver_file = await self.resolver_file()
        command = (
            "massdns",
            "-r",
            resolver_file,
            "-s",
            self.max_resolvers,
            "-t",
            rdtype,
            "-o",
            "J",
            "-q",
        )
        subdomains = self.gen_subdomains(subdomains, domain)
        hosts_yielded = set()
        async with self._dnsbrute_lock:
            async for line in module.run_process_live(*command, stderr=subprocess.DEVNULL, input=subdomains):
                try:
                    j = json.loads(line)
                except json.decoder.JSONDecodeError:
                    self.log.debug(f"Failed to decode line: {line}")
                    continue
                answers = j.get("data", {}).get("answers", [])
                if type(answers) == list and len(answers) > 0:
                    answer = answers[0]
                    hostname = answer.get("name", "").strip(".").lower()
                    if hostname.endswith(f".{domain}"):
                        data = answer.get("data", "")
                        rdtype = answer.get("type", "").upper()
                        if data and rdtype:
                            hostname_hash = hash(hostname)
                            if hostname_hash not in hosts_yielded:
                                hosts_yielded.add(hostname_hash)
                                yield hostname, data, rdtype

    async def gen_subdomains(self, prefixes, domain):
        for p in prefixes:
            if domain:
                p = f"{p}.{domain}"
            yield p

    async def resolver_file(self):
        if self._resolver_file is None:
            self._resolver_file_original = await self.parent_helper.wordlist(
                self.nameservers_url,
                cache_hrs=24 * 7,
            )
            nameservers = set(self.parent_helper.read_file(self._resolver_file_original))
            nameservers.difference_update(self.parent_helper.dns.system_resolvers)
            # exclude system nameservers from brute-force
            # this helps prevent rate-limiting which might cause BBOT's main dns queries to fail
            self._resolver_file = self.parent_helper.tempfile(nameservers, pipe=False)
        return self._resolver_file

    def gen_random_subdomains(self, n=50):
        delimiters = (".", "-")
        lengths = list(range(3, 8))
        for i in range(0, max(0, n - 5)):
            d = delimiters[i % len(delimiters)]
            l = lengths[i % len(lengths)]
            segments = list(random.choice(self.devops_mutations) for _ in range(l))
            segments.append(self.parent_helper.rand_string(length=8, digits=False))
            subdomain = d.join(segments)
            yield subdomain
        for _ in range(5):
            yield self.parent_helper.rand_string(length=8, digits=False)

    def has_excessive_digits(self, d):
        """
        Identifies dns names with excessive numbers, e.g.:
            - w1-2-3.evilcorp.com
            - ptr1234.evilcorp.com
        """
        is_ptr = self.parent_helper.is_ptr(d)
        digits = self.digit_regex.findall(d)
        excessive_digits = len(digits) > 2
        long_digits = any(len(d) > 3 for d in digits)
        return is_ptr or excessive_digits or long_digits



import regex as re

from .emailformat import emailformat


class skymem(emailformat):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS"]
    flags = ["passive", "email-enum", "safe"]
    meta = {
        "description": "Query skymem.info for email addresses",
        "created_date": "2022-07-11",
        "author": "@TheTechromancer",
    }

    base_url = "https://www.skymem.info"
    _qsize = 1

    async def setup(self):
        self.next_page_regex = self.helpers.re.compile(r'<a href="/domain/([a-z0-9]+)\?p=', re.I)
        return True

    async def handle_event(self, event):
        _, query = self.helpers.split_domain(event.data)
        # get first page
        url = f"{self.base_url}/srch?q={self.helpers.quote(query)}"
        r = await self.request_with_fail_count(url)
        if not r:
            return
        responses = [r]

        # iterate through other pages
        domain_ids = await self.helpers.re.findall(self.next_page_regex, r.text)
        if domain_ids:
            domain_id = domain_ids[0]
            for page in range(2, 22):
                r2 = await self.request_with_fail_count(f"{self.base_url}/domain/{domain_id}?p={page}")
                if not r2:
                    continue
                responses.append(r2)
                pages = re.findall(r"/domain/" + domain_id + r"\?p=(\d+)", r2.text)
                if not pages:
                    break
                last_page = max([int(p) for p in pages])
                if page >= last_page:
                    break

        for i, r in enumerate(responses):
            for email in await self.helpers.re.extract_emails(r.text):
                await self.emit_event(
                    email,
                    "EMAIL_ADDRESS",
                    parent=event,
                    context=f'{{module}} searched skymem.info for "{query}" and found {{event.type}} on page {i+1}: {{event.data}}',
                )



from baddns.base import get_all_modules
from .baddns import baddns as baddns_module


class baddns_zone(baddns_module):
    watched_events = ["DNS_NAME"]
    produced_events = ["FINDING", "VULNERABILITY"]
    flags = ["active", "safe", "subdomain-enum", "baddns", "cloud-enum"]
    meta = {
        "description": "Check hosts for DNS zone transfers and NSEC walks",
        "created_date": "2024-01-29",
        "author": "@liquidsec",
    }
    options = {"custom_nameservers": [], "only_high_confidence": False}
    options_desc = {
        "custom_nameservers": "Force BadDNS to use a list of custom nameservers",
        "only_high_confidence": "Do not emit low-confidence or generic detections",
    }
    module_threads = 8
    deps_pip = ["baddns~=1.1.815"]

    def select_modules(self):
        selected_modules = []
        for m in get_all_modules():
            if m.name in ["NSEC", "zonetransfer"]:
                selected_modules.append(m)
        return selected_modules

    # minimize nsec records feeding back into themselves
    async def filter_event(self, event):
        if "baddns-nsec" in event.tags or "baddns-nsec" in event.parent.tags:
            return False
        return True



from .base import ModuleTestBase


class TestAffiliates(ModuleTestBase):
    targets = ["8.8.8.8"]
    config_overrides = {"dns": {"minimal": False}}

    async def setup_before_prep(self, module_test):
        await module_test.mock_dns(
            {
                "8.8.8.8.in-addr.arpa": {"PTR": ["dns.google"]},
                "dns.google": {"A": ["8.8.8.8"], "NS": ["ns1.zdns.google"]},
                "ns1.zdns.google": {"A": ["1.2.3.4"]},
            }
        )

    def check(self, module_test, events):
        filename = next(module_test.scan.home.glob("affiliates-table*.txt"))
        with open(filename) as f:
            assert "zdns.google" in f.read()



from bbot.errors import WebError
from bbot.modules.output.base import BaseOutputModule


class HTTP(BaseOutputModule):
    watched_events = ["*"]
    meta = {
        "description": "Send every event to a custom URL via a web request",
        "created_date": "2022-04-13",
        "author": "@TheTechromancer",
    }
    options = {
        "url": "",
        "method": "POST",
        "bearer": "",
        "username": "",
        "password": "",
        "timeout": 10,
        "siem_friendly": False,
    }
    options_desc = {
        "url": "Web URL",
        "method": "HTTP method",
        "bearer": "Authorization Bearer token",
        "username": "Username (basic auth)",
        "password": "Password (basic auth)",
        "timeout": "HTTP timeout",
        "siem_friendly": "Format JSON in a SIEM-friendly way for ingestion into Elastic, Splunk, etc.",
    }

    async def setup(self):
        self.url = self.config.get("url", "")
        self.method = self.config.get("method", "POST")
        self.timeout = self.config.get("timeout", 10)
        self.siem_friendly = self.config.get("siem_friendly", False)
        self.headers = {}
        bearer = self.config.get("bearer", "")
        if bearer:
            self.headers["Authorization"] = f"Bearer {bearer}"
        username = self.config.get("username", "")
        password = self.config.get("password", "")
        self.auth = None
        if username:
            self.auth = (username, password)
        if not self.url:
            self.warning("Must set URL")
            return False
        if not self.method:
            self.warning("Must set HTTP method")
            return False
        return True

    async def handle_event(self, event):
        while 1:
            try:
                await self.helpers.request(
                    url=self.url,
                    method=self.method,
                    auth=self.auth,
                    headers=self.headers,
                    json=event.json(siem_friendly=self.siem_friendly),
                    raise_error=True,
                )
                break
            except WebError as e:
                self.warning(f"Error sending {event}: {e}, retrying...")
                await self.helpers.sleep(1)



import re

from ..bbot_fixtures import *

from bbot.modules.base import BaseModule
from bbot.modules.output.base import BaseOutputModule
from bbot.modules.report.base import BaseReportModule
from bbot.modules.internal.base import BaseInternalModule


@pytest.mark.asyncio
async def test_modules_basic_checks(events, httpx_mock):
    for http_method in ("GET", "CONNECT", "HEAD", "POST", "PUT", "TRACE", "DEBUG", "PATCH", "DELETE", "OPTIONS"):
        httpx_mock.add_response(method=http_method, url=re.compile(r".*"), json={"test": "test"})

    from bbot.scanner import Scanner

    scan = Scanner(config={"omit_event_types": ["URL_UNVERIFIED"]})
    assert "URL_UNVERIFIED" in scan.omitted_event_types

    await scan.load_modules()

    # output module specific event filtering tests
    base_output_module_1 = BaseOutputModule(scan)
    base_output_module_1.watched_events = ["IP_ADDRESS", "URL_UNVERIFIED"]
    localhost = scan.make_event("127.0.0.1", parent=scan.root_event)
    # ip addresses should be accepted
    result, reason = base_output_module_1._event_precheck(localhost)
    assert result == True
    assert reason == "precheck succeeded"
    # internal events should be rejected
    localhost._internal = True
    result, reason = base_output_module_1._event_precheck(localhost)
    assert result == False
    assert reason == "_internal is True"
    localhost._internal = False
    result, reason = base_output_module_1._event_precheck(localhost)
    assert result == True
    assert reason == "precheck succeeded"
    # unwatched events should be rejected
    dns_name = scan.make_event("evilcorp.com", parent=scan.root_event)
    result, reason = base_output_module_1._event_precheck(dns_name)
    assert result == False
    assert reason == "its type is not in watched_events"
    # omitted events matching watched types should be accepted
    url_unverified = scan.make_event("http://127.0.0.1", "URL_UNVERIFIED", parent=scan.root_event)
    url_unverified._omit = True
    result, reason = base_output_module_1._event_precheck(url_unverified)
    assert result == True
    assert reason == "its type is explicitly in watched_events"

    base_output_module_2 = BaseOutputModule(scan)
    base_output_module_2.watched_events = ["*"]
    # normal events should be accepted
    localhost = scan.make_event("127.0.0.1", parent=scan.root_event)
    result, reason = base_output_module_2._event_precheck(localhost)
    assert result == True
    assert reason == "precheck succeeded"
    # internal events should be rejected
    localhost._internal = True
    result, reason = base_output_module_2._event_precheck(localhost)
    assert result == False
    assert reason == "_internal is True"
    localhost._internal = False
    result, reason = base_output_module_2._event_precheck(localhost)
    assert result == True
    assert reason == "precheck succeeded"
    # omitted events should be rejected
    localhost._omit = True
    result, reason = base_output_module_2._event_precheck(localhost)
    assert result == False
    assert reason == "_omit is True"
    # normal event should be accepted
    url_unverified = scan.make_event("http://127.0.0.1", "URL_UNVERIFIED", parent=scan.root_event)
    result, reason = base_output_module_2._event_precheck(url_unverified)
    assert result == True
    assert reason == "precheck succeeded"
    # omitted event types should be marked during scan egress
    await scan.egress_module.handle_event(url_unverified)
    result, reason = base_output_module_2._event_precheck(url_unverified)
    assert result == False
    assert reason == "_omit is True"
    # omitted events that are targets should be accepted
    dns_name = scan.make_event("evilcorp.com", "DNS_NAME", parent=scan.root_event)
    dns_name._omit = True
    result, reason = base_output_module_2._event_precheck(dns_name)
    assert result == False
    assert reason == "_omit is True"
    # omitted results that are targets should be accepted
    dns_name.add_tag("target")
    result, reason = base_output_module_2._event_precheck(dns_name)
    assert result == True
    assert reason == "it's a target"

    # common event filtering tests
    for module_class in (BaseModule, BaseOutputModule, BaseReportModule, BaseInternalModule):
        base_module = module_class(scan)
        localhost2 = scan.make_event("127.0.0.2", parent=events.subdomain)
        localhost2.scope_distance = 0
        # base cases
        base_module._watched_events = None
        base_module.watched_events = ["*"]
        assert base_module._event_precheck(events.emoji)[0] == True
        base_module._watched_events = None
        base_module.watched_events = ["IP_ADDRESS"]
        assert base_module._event_precheck(events.ipv4)[0] == True
        assert base_module._event_precheck(events.domain)[0] == False
        assert base_module._event_precheck(events.localhost)[0] == True
        assert base_module._event_precheck(localhost2)[0] == True
        # target only
        base_module.target_only = True
        assert base_module._event_precheck(localhost2)[0] == False
        localhost2.add_tag("target")
        assert base_module._event_precheck(localhost2)[0] == True
        base_module.target_only = False

        # in scope only
        base_module.in_scope_only = True
        localhost3 = scan.make_event("127.0.0.2", parent=events.subdomain)
        valid, reason = await base_module._event_postcheck(localhost3)
        if base_module._type == "output":
            assert valid
        else:
            assert not valid
            assert reason == "it did not meet in_scope_only filter criteria"
        base_module.in_scope_only = False
        base_module.scope_distance_modifier = 0
        valid, reason = await base_module._event_postcheck(events.localhost)
        assert valid

    # module preloading
    all_preloaded = DEFAULT_PRESET.module_loader.preloaded()
    assert "dnsbrute" in all_preloaded
    assert "DNS_NAME" in all_preloaded["dnsbrute"]["watched_events"]
    assert "DNS_NAME" in all_preloaded["dnsbrute"]["produced_events"]
    assert "subdomain-enum" in all_preloaded["dnsbrute"]["flags"]
    assert "wordlist" in all_preloaded["dnsbrute"]["config"]
    assert type(all_preloaded["dnsbrute"]["config"]["max_depth"]) == int
    assert all_preloaded["sslcert"]["deps"]["pip"]
    assert all_preloaded["sslcert"]["deps"]["apt"]
    assert all_preloaded["dnsbrute"]["deps"]["common"]
    assert all_preloaded["gowitness"]["deps"]["ansible"]

    all_flags = set()

    created_date_regex = re.compile(r"2[\d]{3}-[\d]{2}-[\d]{2}")
    for module_name, preloaded in all_preloaded.items():
        # either active or passive and never both
        flags = preloaded.get("flags", [])
        for flag in flags:
            all_flags.add(flag)
        if preloaded["type"] == "scan":
            assert ("active" in flags and not "passive" in flags) or (
                not "active" in flags and "passive" in flags
            ), f'module "{module_name}" must have either "active" or "passive" flag'
            assert ("safe" in flags and not "aggressive" in flags) or (
                not "safe" in flags and "aggressive" in flags
            ), f'module "{module_name}" must have either "safe" or "aggressive" flag'
            assert not (
                "web-basic" in flags and "web-thorough" in flags
            ), f'module "{module_name}" should have either "web-basic" or "web-thorough" flags, not both'
            meta = preloaded.get("meta", {})
            # make sure every module has a description
            assert meta.get("description", ""), f"{module_name} must have a description"
            # make sure every module has an author
            assert meta.get("author", ""), f"{module_name} must have an author"
            # make sure every module has a created date
            created_date = meta.get("created_date", "")
            assert created_date, f"{module_name} must have a created date"
            assert created_date_regex.match(
                created_date
            ), f"{module_name}'s created_date must match the format YYYY-MM-DD"

        # attribute checks
        watched_events = preloaded.get("watched_events")
        produced_events = preloaded.get("produced_events")

        assert type(watched_events) == list
        assert type(produced_events) == list
        if not preloaded.get("type", "") in ("internal",):
            assert watched_events, f"{module_name}.watched_events must not be empty"
        assert type(watched_events) == list, f"{module_name}.watched_events must be of type list"
        assert type(produced_events) == list, f"{module_name}.produced_events must be of type list"
        assert all(
            [type(t) == str for t in watched_events]
        ), f"{module_name}.watched_events entries must be of type string"
        assert all(
            [type(t) == str for t in produced_events]
        ), f"{module_name}.produced_events entries must be of type string"

        assert type(preloaded.get("deps_pip", [])) == list, f"{module_name}.deps_pip must be of type list"
        assert (
            type(preloaded.get("deps_pip_constraints", [])) == list
        ), f"{module_name}.deps_pip_constraints must be of type list"
        assert type(preloaded.get("deps_apt", [])) == list, f"{module_name}.deps_apt must be of type list"
        assert type(preloaded.get("deps_shell", [])) == list, f"{module_name}.deps_shell must be of type list"
        assert type(preloaded.get("config", None)) == dict, f"{module_name}.options must be of type list"
        assert type(preloaded.get("options_desc", None)) == dict, f"{module_name}.options_desc must be of type list"
        # options must have descriptions
        assert set(preloaded.get("config", {})) == set(
            preloaded.get("options_desc", {})
        ), f"{module_name}.options do not match options_desc"
        # descriptions most not be blank
        assert all(
            o for o in preloaded.get("options_desc", {}).values()
        ), f"{module_name}.options_desc descriptions must not be blank"

    from bbot.core.flags import flag_descriptions

    for flag in all_flags:
        assert flag in flag_descriptions, f'Flag "{flag}" not listed in bbot/core/flags.py'
        description = flag_descriptions.get(flag, "")
        assert description, f'Flag "{flag}" has no description in bbot/core/flags.py'

    await scan._cleanup()


@pytest.mark.asyncio
async def test_modules_basic_perhostonly(bbot_scanner):
    from bbot.modules.base import BaseModule

    class mod_normal(BaseModule):
        _name = "mod_normal"
        watched_events = ["*"]

    class mod_host_only(BaseModule):
        _name = "mod_hostonly"
        watched_events = ["*"]
        per_host_only = True

    class mod_hostport_only(BaseModule):
        _name = "mod_normal"
        watched_events = ["*"]
        per_hostport_only = True

    class mod_domain_only(BaseModule):
        _name = "domain_only"
        watched_events = ["*"]
        per_domain_only = True

    scan = bbot_scanner(
        "evilcorp.com",
        force_start=True,
    )

    scan.modules["mod_normal"] = mod_normal(scan)
    scan.modules["mod_host_only"] = mod_host_only(scan)
    scan.modules["mod_hostport_only"] = mod_hostport_only(scan)
    scan.modules["mod_domain_only"] = mod_domain_only(scan)
    scan.status = "RUNNING"

    url_1 = scan.make_event("http://evilcorp.com/1", event_type="URL", parent=scan.root_event, tags=["status-200"])
    url_2 = scan.make_event("http://evilcorp.com/2", event_type="URL", parent=scan.root_event, tags=["status-200"])
    url_3 = scan.make_event("http://evilcorp.com:888/3", event_type="URL", parent=scan.root_event, tags=["status-200"])
    url_4 = scan.make_event("http://www.evilcorp.com/", event_type="URL", parent=scan.root_event, tags=["status-200"])
    url_5 = scan.make_event("http://www.evilcorp.net/", event_type="URL", parent=scan.root_event, tags=["status-200"])

    url_1.scope_distance = 0
    url_2.scope_distance = 0
    url_3.scope_distance = 0
    url_4.scope_distance = 0
    url_5.scope_distance = 0

    for mod_name in ("mod_normal", "mod_host_only", "mod_hostport_only", "mod_domain_only"):
        module = scan.modules[mod_name]

        valid_1, reason_1 = await module._event_postcheck(url_1)
        valid_2, reason_2 = await module._event_postcheck(url_2)
        valid_3, reason_3 = await module._event_postcheck(url_3)
        valid_4, reason_4 = await module._event_postcheck(url_4)
        valid_5, reason_5 = await module._event_postcheck(url_5)

        if mod_name == "mod_normal":
            assert valid_1 == True
            assert valid_2 == True
            assert valid_3 == True
            assert valid_4 == True
            assert valid_5 == True
        elif mod_name == "mod_host_only":
            assert valid_1 == True
            assert valid_2 == False
            assert "per_host_only=True" in reason_2
            assert valid_3 == False
            assert "per_host_only=True" in reason_3
            assert valid_4 == True
            assert valid_5 == True
        elif mod_name == "mod_hostport_only":
            assert valid_1 == True
            assert valid_2 == False
            assert "per_hostport_only=True" in reason_2
            assert valid_3 == True
            assert valid_4 == True
            assert valid_5 == True
        elif mod_name == "mod_domain_only":
            assert valid_1 == True
            assert valid_2 == False
            assert "per_domain_only=True" in reason_2
            assert valid_3 == False
            assert "per_domain_only=True" in reason_3
            assert valid_4 == False
            assert "per_domain_only=True" in reason_4
            assert valid_5 == True

    await scan._cleanup()


@pytest.mark.asyncio
async def test_modules_basic_perdomainonly(bbot_scanner, monkeypatch):
    per_domain_scan = bbot_scanner(
        "evilcorp.com",
        modules=list(available_modules),
        config={i: True for i in available_internal_modules if i != "dnsresolve"},
        force_start=True,
    )

    await per_domain_scan.load_modules()
    await per_domain_scan.setup_modules()
    per_domain_scan.status = "RUNNING"

    # ensure that multiple events to the same "host" (schema + host) are blocked and check the per host tracker

    for module_name, module in sorted(per_domain_scan.modules.items()):
        monkeypatch.setattr(module, "filter_event", BaseModule(per_domain_scan).filter_event)

        if "URL" in module.watched_events:
            url_1 = per_domain_scan.make_event(
                "http://www.evilcorp.com/1", event_type="URL", parent=per_domain_scan.root_event, tags=["status-200"]
            )
            url_1.scope_distance = 0
            url_2 = per_domain_scan.make_event(
                "http://mail.evilcorp.com/2", event_type="URL", parent=per_domain_scan.root_event, tags=["status-200"]
            )
            url_2.scope_distance = 0
            valid_1, reason_1 = await module._event_postcheck(url_1)
            valid_2, reason_2 = await module._event_postcheck(url_2)

            if module.per_domain_only == True:
                assert valid_1 == True
                assert valid_2 == False
                assert hash("evilcorp.com") in module._per_host_tracker
                assert reason_2 == "per_domain_only enabled and already seen domain"

            else:
                assert valid_1 == True
                assert valid_2 == True

    await per_domain_scan._cleanup()


@pytest.mark.asyncio
async def test_modules_basic_stats(helpers, events, bbot_scanner, httpx_mock, monkeypatch):
    from bbot.modules.base import BaseModule

    class dummy(BaseModule):
        _name = "dummy"
        watched_events = ["*"]

        async def handle_event(self, event):
            # quick emit events like FINDINGS behave differently than normal ones
            # hosts are not speculated from them
            await self.emit_event(
                {"host": "www.evilcorp.com", "url": "http://www.evilcorp.com", "description": "asdf"}, "FINDING", event
            )
            await self.emit_event("https://asdf.evilcorp.com", "URL", event, tags=["status-200"])

    scan = bbot_scanner(
        "evilcorp.com",
        config={"speculate": True, "dns": {"minimal": False}},
        output_modules=["python"],
        force_start=True,
    )
    await scan.helpers.dns._mock_dns(
        {
            "evilcorp.com": {"A": ["127.0.254.1"]},
            "www.evilcorp.com": {"A": ["127.0.254.2"]},
            "asdf.evilcorp.com": {"A": ["127.0.254.3"]},
        },
    )

    scan.modules["dummy"] = dummy(scan)
    events = [e async for e in scan.async_start()]

    assert len(events) == 8
    assert 1 == len([e for e in events if e.type == "SCAN"])
    assert 3 == len([e for e in events if e.type == "DNS_NAME"])
    # one from target and one from speculate
    assert 2 == len([e for e in events if e.type == "DNS_NAME" and e.data == "evilcorp.com"])
    # the reason we don't have a DNS_NAME for www.evilcorp.com is because FINDING.quick_emit = True
    assert 0 == len([e for e in events if e.type == "DNS_NAME" and e.data == "www.evilcorp.com"])
    assert 1 == len([e for e in events if e.type == "DNS_NAME" and e.data == "asdf.evilcorp.com"])
    assert 1 == len([e for e in events if e.type == "ORG_STUB" and e.data == "evilcorp"])
    assert 1 == len([e for e in events if e.type == "FINDING"])
    assert 1 == len([e for e in events if e.type == "URL_UNVERIFIED"])

    assert scan.stats.events_emitted_by_type == {
        "SCAN": 1,
        "DNS_NAME": 3,
        "URL": 1,
        "ORG_STUB": 1,
        "URL_UNVERIFIED": 1,
        "FINDING": 1,
        "ORG_STUB": 1,
    }

    assert set(scan.stats.module_stats) == {"speculate", "host", "TARGET", "python", "dummy", "dnsresolve"}

    target_stats = scan.stats.module_stats["TARGET"]
    assert target_stats.produced == {"SCAN": 1, "DNS_NAME": 1}
    assert target_stats.produced_total == 2
    assert target_stats.consumed == {}
    assert target_stats.consumed_total == 0

    dummy_stats = scan.stats.module_stats["dummy"]
    assert dummy_stats.produced == {"FINDING": 1, "URL": 1}
    assert dummy_stats.produced_total == 2
    assert dummy_stats.consumed == {
        "DNS_NAME": 2,
        "FINDING": 1,
        "OPEN_TCP_PORT": 1,
        "ORG_STUB": 1,
        "SCAN": 1,
        "URL": 1,
        "URL_UNVERIFIED": 1,
    }
    assert dummy_stats.consumed_total == 8

    python_stats = scan.stats.module_stats["python"]
    assert python_stats.produced == {}
    assert python_stats.produced_total == 0
    assert python_stats.consumed == {
        "DNS_NAME": 3,
        "FINDING": 1,
        "ORG_STUB": 1,
        "SCAN": 1,
        "URL": 1,
        "URL_UNVERIFIED": 1,
    }
    assert python_stats.consumed_total == 8

    speculate_stats = scan.stats.module_stats["speculate"]
    assert speculate_stats.produced == {"DNS_NAME": 1, "URL_UNVERIFIED": 1, "ORG_STUB": 1}
    assert speculate_stats.produced_total == 3
    assert speculate_stats.consumed == {"URL": 1, "DNS_NAME": 2, "URL_UNVERIFIED": 1, "IP_ADDRESS": 2}
    assert speculate_stats.consumed_total == 6


@pytest.mark.asyncio
async def test_module_loading(bbot_scanner):
    scan2 = bbot_scanner(
        modules=list(available_modules),
        output_modules=list(available_output_modules),
        config={i: True for i in available_internal_modules if i != "dnsresolve"},
        force_start=True,
    )
    await scan2.load_modules()
    scan2.status = "RUNNING"

    # attributes, descriptions, etc.
    for module_name, module in sorted(scan2.modules.items()):
        # flags
        assert module._type in ("internal", "output", "scan")
        # async stuff
        not_async = []
        for func_name in ("setup", "ping", "filter_event", "handle_event", "finish", "report", "cleanup"):
            f = getattr(module, func_name)
            if not scan2.helpers.is_async_function(f):
                log.error(f"{f.__qualname__}() is not async")
                not_async.append(f)
    assert not any(not_async)

    await scan2._cleanup()



from bbot.modules.base import BaseModule


class bucket_template(BaseModule):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic"]
    options = {"permutations": False}
    options_desc = {
        "permutations": "Whether to try permutations",
    }
    scope_distance_modifier = 3

    cloud_helper_name = "amazon|google|digitalocean|etc"
    delimiters = ("", ".", "-")
    base_domains = ["s3.amazonaws.com|digitaloceanspaces.com|etc"]
    regions = [None]
    supports_open_check = True

    async def setup(self):
        self.buckets_tried = set()
        self.cloud_helper = self.helpers.cloud.providers[self.cloud_helper_name]
        self.permutations = self.config.get("permutations", False)
        return True

    async def filter_event(self, event):
        if event.type == "DNS_NAME" and event.scope_distance > 0:
            return False, "only accepts in-scope DNS_NAMEs"
        if event.type == "STORAGE_BUCKET":
            filter_result, reason = self.filter_bucket(event)
            if not filter_result:
                return (filter_result, reason)
        return True

    def filter_bucket(self, event):
        if f"cloud-{self.cloud_helper_name}" not in event.tags:
            return False, "bucket belongs to a different cloud provider"
        return True, ""

    async def handle_event(self, event):
        if event.type == "DNS_NAME":
            await self.handle_dns_name(event)
        elif event.type == "STORAGE_BUCKET":
            await self.handle_storage_bucket(event)

    async def handle_dns_name(self, event):
        buckets = set()
        base = self.helpers.unidecode(self.helpers.smart_decode_punycode(event.data))
        stem = self.helpers.domain_stem(base)
        for b in [base, stem]:
            split = b.split(".")
            for d in self.delimiters:
                bucket_name = d.join(split)
                buckets.add(bucket_name)
        async for bucket_name, url, tags, num_buckets in self.brute_buckets(buckets, permutations=self.permutations):
            await self.emit_storage_bucket(
                {"name": bucket_name, "url": url},
                "STORAGE_BUCKET",
                parent=event,
                tags=tags,
                context=f"{{module}} tried {num_buckets:,} bucket variations of {event.data} and found {{event.type}} at {url}",
            )

    async def handle_storage_bucket(self, event):
        url = event.data["url"]
        bucket_name = event.data["name"]
        if self.supports_open_check:
            description, tags = await self._check_bucket_open(bucket_name, url)
            if description:
                event_data = {"host": event.host, "url": url, "description": description}
                await self.emit_event(
                    event_data,
                    "FINDING",
                    parent=event,
                    tags=tags,
                    context=f"{{module}} scanned {event.type} and identified {{event.type}}: {description}",
                )

        async for bucket_name, new_url, tags, num_buckets in self.brute_buckets(
            [bucket_name], permutations=self.permutations, omit_base=True
        ):
            await self.emit_storage_bucket(
                {"name": bucket_name, "url": new_url},
                "STORAGE_BUCKET",
                parent=event,
                tags=tags,
                context=f"{{module}} tried {num_buckets:,} variations of {url} and found {{event.type}} at {new_url}",
            )

    async def emit_storage_bucket(self, event_data, event_type, parent, tags, context):
        event_data["url"] = self.clean_bucket_url(event_data["url"])
        await self.emit_event(
            event_data,
            event_type,
            parent=parent,
            tags=tags,
            context=context,
        )

    async def brute_buckets(self, buckets, permutations=False, omit_base=False):
        buckets = set(buckets)
        new_buckets = set(buckets)
        if permutations:
            for b in buckets:
                for mutation in self.helpers.word_cloud.mutations(b, cloud=False):
                    for d in self.delimiters:
                        new_buckets.add(d.join(mutation))
        if omit_base:
            new_buckets = new_buckets - buckets
        new_buckets = [b for b in new_buckets if self.valid_bucket_name(b)]
        num_buckets = len(new_buckets)
        bucket_urls_kwargs = []
        for base_domain in self.base_domains:
            for region in self.regions:
                for bucket_name in new_buckets:
                    url, kwargs = self.build_bucket_request(bucket_name, base_domain, region)
                    bucket_urls_kwargs.append((url, kwargs, (bucket_name, base_domain, region)))
        async for url, kwargs, (bucket_name, base_domain, region), response in self.helpers.request_custom_batch(
            bucket_urls_kwargs
        ):
            existent_bucket, tags = self._check_bucket_exists(bucket_name, response)
            if existent_bucket:
                yield bucket_name, url, tags, num_buckets

    def clean_bucket_url(self, url):
        # if needed, modify the bucket url before emitting it
        return url

    def build_bucket_request(self, bucket_name, base_domain, region):
        url = self.build_url(bucket_name, base_domain, region)
        return url, {}

    def _check_bucket_exists(self, bucket_name, response):
        self.debug(f'Checking if bucket exists: "{bucket_name}"')
        return self.check_bucket_exists(bucket_name, response)

    def check_bucket_exists(self, bucket_name, response):
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        existent_bucket = status_code != 404
        return (existent_bucket, tags)

    async def _check_bucket_open(self, bucket_name, url):
        self.debug(f'Checking if bucket is misconfigured: "{bucket_name}"')
        return await self.check_bucket_open(bucket_name, url)

    async def check_bucket_open(self, bucket_name, url):
        response = await self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        content = getattr(response, "text", "")
        open_bucket = status_code == 200 and "Contents" in content
        msg = ""
        if open_bucket:
            msg = "Open storage bucket"
        return (msg, tags)

    def valid_bucket_name(self, bucket_name):
        valid = self.cloud_helper.is_valid_bucket_name(bucket_name)
        if valid and not self.helpers.is_ip(bucket_name):
            bucket_hash = hash(bucket_name)
            if not bucket_hash in self.buckets_tried:
                self.buckets_tried.add(bucket_hash)
                return True
        return False

    def build_url(self, bucket_name, base_domain, region):
        return f"https://{bucket_name}.{base_domain}/"

    def gen_tags_exists(self, response):
        return set()

    def gen_tags_open(self, response):
        return set()



import re
import random
import string

from bbot.modules.deadly.ffuf import ffuf


def find_common_prefixes(strings, minimum_set_length=4):
    prefix_candidates = [s[:i] for s in strings if len(s) == 6 for i in range(3, 6)]
    frequency_dict = {item: prefix_candidates.count(item) for item in prefix_candidates}
    frequency_dict = {k: v for k, v in frequency_dict.items() if v >= minimum_set_length}
    prefix_list = list(set(frequency_dict.keys()))

    found_prefixes = set()
    for prefix in prefix_list:
        prefix_frequency = frequency_dict[prefix]
        is_substring = False

        for k, v in frequency_dict.items():
            if prefix != k:
                if prefix in k:
                    is_substring = True
        if not is_substring:
            found_prefixes.add(prefix)
        else:
            if prefix_frequency > v and (len(k) - len(prefix) == 1):
                found_prefixes.add(prefix)
    return list(found_prefixes)


class ffuf_shortnames(ffuf):
    watched_events = ["URL_HINT"]
    produced_events = ["URL_UNVERIFIED"]
    flags = ["aggressive", "active", "iis-shortnames", "web-thorough"]
    meta = {
        "description": "Use ffuf in combination IIS shortnames",
        "created_date": "2022-07-05",
        "author": "@liquidsec",
    }

    options = {
        "wordlist": "",  # default is defined within setup function
        "wordlist_extensions": "",  # default is defined within setup function
        "lines": 1000000,
        "max_depth": 1,
        "version": "2.0.0",
        "extensions": "",
        "ignore_redirects": True,
        "find_common_prefixes": False,
        "find_delimiters": True,
    }

    options_desc = {
        "wordlist": "Specify wordlist to use when finding directories",
        "wordlist_extensions": "Specify wordlist to use when making extension lists",
        "lines": "take only the first N lines from the wordlist when finding directories",
        "max_depth": "the maximum directory depth to attempt to solve",
        "version": "ffuf version",
        "extensions": "Optionally include a list of extensions to extend the keyword with (comma separated)",
        "ignore_redirects": "Explicitly ignore redirects (301,302)",
        "find_common_prefixes": "Attempt to automatically detect common prefixes and make additional ffuf runs against them",
        "find_delimiters": "Attempt to detect common delimiters and make additional ffuf runs against them",
    }

    deps_common = ["ffuf"]

    in_scope_only = True

    async def setup(self):
        self.proxy = self.scan.web_config.get("http_proxy", "")
        self.canary = "".join(random.choice(string.ascii_lowercase) for i in range(10))
        wordlist = self.config.get("wordlist", "")
        if not wordlist:
            wordlist = f"{self.helpers.wordlist_dir}/ffuf_shortname_candidates.txt"
        self.debug(f"Using [{wordlist}] for shortname candidate list")
        self.wordlist = await self.helpers.wordlist(wordlist)
        self.wordlist_lines = self.generate_wordlist(self.wordlist)

        wordlist_extensions = self.config.get("wordlist_extensions", "")
        if not wordlist_extensions:
            wordlist_extensions = f"{self.helpers.wordlist_dir}/raft-small-extensions-lowercase_CLEANED.txt"
        self.debug(f"Using [{wordlist_extensions}] for shortname candidate extension list")
        self.wordlist_extensions = await self.helpers.wordlist(wordlist_extensions)

        try:
            self.extensions = self.helpers.chain_lists(self.config.get("extensions", ""), validate=True)
            self.debug(f"Using custom extensions: [{','.join(self.extensions)}]")
        except ValueError as e:
            return False, f"Error parsing extensions: {e}"

        self.ignore_redirects = self.config.get("ignore_redirects")

        self.per_host_collection = {}
        self.shortname_to_event = {}
        return True

    def build_extension_list(self, event):
        used_extensions = []
        extension_hint = event.parsed_url.path.rsplit(".", 1)[1].lower().strip()
        if len(extension_hint) == 3:
            with open(self.wordlist_extensions) as f:
                for l in f:
                    l = l.lower().lstrip(".")
                    if l.lower().startswith(extension_hint):
                        used_extensions.append(l.strip())
            return used_extensions
        else:
            return [extension_hint]

    def find_delimiter(self, hint):
        delimiters = ["_", "-"]
        for d in delimiters:
            if d in hint:
                if not hint.startswith(d) and not hint.endswith(d):
                    return d, hint.split(d)[0], hint.split(d)[1]
        return None

    async def filter_event(self, event):
        if event.parent.type != "URL":
            return False, "its parent event is not of type URL"
        return True

    async def handle_event(self, event):
        filename_hint = re.sub(r"~\d", "", event.parsed_url.path.rsplit(".", 1)[0].split("/")[-1]).lower()

        host = f"{event.parent.parsed_url.scheme}://{event.parent.parsed_url.netloc}/"
        if host not in self.per_host_collection.keys():
            self.per_host_collection[host] = [(filename_hint, event.parent.data)]

        else:
            self.per_host_collection[host].append((filename_hint, event.parent.data))

        self.shortname_to_event[filename_hint] = event

        root_stub = "/".join(event.parsed_url.path.split("/")[:-1])
        root_url = f"{event.parsed_url.scheme}://{event.parsed_url.netloc}{root_stub}/"

        if "shortname-file" in event.tags:
            used_extensions = self.build_extension_list(event)

        if len(filename_hint) == 6:
            tempfile, tempfile_len = self.generate_templist(prefix=filename_hint)
            self.verbose(
                f"generated temp word list of size [{str(tempfile_len)}] for filename hint: [{filename_hint}]"
            )

        else:
            tempfile = self.helpers.tempfile([filename_hint], pipe=False)
            tempfile_len = 1

        if tempfile_len > 0:
            if "shortname-file" in event.tags:
                for ext in used_extensions:
                    async for r in self.execute_ffuf(tempfile, root_url, suffix=f".{ext}"):
                        await self.emit_event(
                            r["url"],
                            "URL_UNVERIFIED",
                            parent=event,
                            tags=[f"status-{r['status']}"],
                            context=f"{{module}} brute-forced {ext.upper()} files at {root_url} and found {{event.type}}: {{event.data}}",
                        )

            elif "shortname-directory" in event.tags:
                async for r in self.execute_ffuf(tempfile, root_url, exts=["/"]):
                    r_url = f"{r['url'].rstrip('/')}/"
                    await self.emit_event(
                        r_url,
                        "URL_UNVERIFIED",
                        parent=event,
                        tags=[f"status-{r['status']}"],
                        context=f"{{module}} brute-forced directories at {r_url} and found {{event.type}}: {{event.data}}",
                    )

        if self.config.get("find_delimiters"):
            if "shortname-directory" in event.tags:
                delimiter_r = self.find_delimiter(filename_hint)
                if delimiter_r:
                    delimiter, prefix, partial_hint = delimiter_r
                    self.verbose(f"Detected delimiter [{delimiter}] in hint [{filename_hint}]")
                    tempfile, tempfile_len = self.generate_templist(prefix=partial_hint)
                    ffuf_prefix = f"{prefix}{delimiter}"
                    async for r in self.execute_ffuf(tempfile, root_url, prefix=ffuf_prefix, exts=["/"]):
                        await self.emit_event(
                            r["url"],
                            "URL_UNVERIFIED",
                            parent=event,
                            tags=[f"status-{r['status']}"],
                            context=f'{{module}} brute-forced directories with detected prefix "{ffuf_prefix}" and found {{event.type}}: {{event.data}}',
                        )

            elif "shortname-file" in event.tags:
                for ext in used_extensions:
                    delimiter_r = self.find_delimiter(filename_hint)
                    if delimiter_r:
                        delimiter, prefix, partial_hint = delimiter_r
                        self.verbose(f"Detected delimiter [{delimiter}] in hint [{filename_hint}]")
                        tempfile, tempfile_len = self.generate_templist(prefix=partial_hint)
                        ffuf_prefix = f"{prefix}{delimiter}"
                        async for r in self.execute_ffuf(tempfile, root_url, prefix=ffuf_prefix, suffix=f".{ext}"):
                            await self.emit_event(
                                r["url"],
                                "URL_UNVERIFIED",
                                parent=event,
                                tags=[f"status-{r['status']}"],
                                context=f'{{module}} brute-forced {ext.upper()} files with detected prefix "{ffuf_prefix}" and found {{event.type}}: {{event.data}}',
                            )

    async def finish(self):
        if self.config.get("find_common_prefixes"):
            per_host_collection = dict(self.per_host_collection)
            self.per_host_collection.clear()

            for host, hint_tuple_list in per_host_collection.items():
                hint_list = [x[0] for x in hint_tuple_list]

                common_prefixes = find_common_prefixes(hint_list)
                for prefix in common_prefixes:
                    self.verbose(f"Found common prefix: [{prefix}] for host [{host}]")
                    for hint_tuple in hint_tuple_list:
                        hint, url = hint_tuple
                        if hint.startswith(prefix):
                            partial_hint = hint[len(prefix) :]

                            # safeguard to prevent loading the entire wordlist
                            if len(partial_hint) > 0:
                                tempfile, tempfile_len = self.generate_templist(prefix=partial_hint)

                                if "shortname-directory" in self.shortname_to_event[hint].tags:
                                    self.verbose(
                                        f"Running common prefix check for URL_HINT: {hint} with prefix: {prefix} and partial_hint: {partial_hint}"
                                    )

                                    async for r in self.execute_ffuf(tempfile, url, prefix=prefix, exts=["/"]):
                                        await self.emit_event(
                                            r["url"],
                                            "URL_UNVERIFIED",
                                            parent=self.shortname_to_event[hint],
                                            tags=[f"status-{r['status']}"],
                                            context=f'{{module}} brute-forced directories with common prefix "{prefix}" and found {{event.type}}: {{event.data}}',
                                        )
                                elif "shortname-file" in self.shortname_to_event[hint].tags:
                                    used_extensions = self.build_extension_list(self.shortname_to_event[hint])

                                    for ext in used_extensions:
                                        self.verbose(
                                            f"Running common prefix check for URL_HINT: {hint} with prefix: {prefix}, extension: .{ext}, and partial_hint: {partial_hint}"
                                        )
                                        async for r in self.execute_ffuf(
                                            tempfile, url, prefix=prefix, suffix=f".{ext}"
                                        ):
                                            await self.emit_event(
                                                r["url"],
                                                "URL_UNVERIFIED",
                                                parent=self.shortname_to_event[hint],
                                                tags=[f"status-{r['status']}"],
                                                context=f'{{module}} brute-forced {ext.upper()} files with common prefix "{prefix}" and found {{event.type}}: {{event.data}}',
                                            )
